(* 
 * Implementation for 
 *   JS Code Generation
 * Author: Lixing Dong
 * Email: cee.xingweimian@gmail.com
 *)

open Ast

module StringMap = Map.Make(String)

module Generator =
struct

type tpmen = string * bool

type omap =
{
    mutable map: tpmen list StringMap.t;
}

let objmap = { map = StringMap.empty }


let rec generate (lst: stc list) = 
    match lst with
    | [] -> "\n"
    | head::[] -> generate_stc head ^ "\n"
    | head::tail -> 
        let h = generate_stc head in
        let t = generate tail in
        h ^ "\n" ^ t


and generate_stc (s: stc) = 
    match s with
    | StcOfObjDef o -> generate_obj_def o
    | StcOfFuncDef f -> generate_func_def f
    | StcOfStmt s -> generate_stmt s


and generate_func_def (fdef: func_def) =
    match fdef with
    FuncDefOfFuncDecAndStcList (fdec, tp, lst) ->
        let fdec = generate_func_dec fdec in
        let lst = generate_in_func_stc_list lst in
        fdec ^ "\n" ^
        "{\n" ^
        lst ^
        "}\n"


and generate_func_dec (fdec: func_dec) =
    match fdec with
    FuncDeclaratorOfIDAndParameters (id, lst) ->
    let lst = generate_parameter_list lst in
    "function " ^ id ^ "(" ^ lst ^ ")"


and generate_parameter_list (lst: parameter list) =
    match lst with
    | [] -> ""
    | head :: [] -> generate_parameter head
    | head :: tail -> 
        let head = generate_parameter head in
        let tail = generate_parameter_list tail in
        head ^ ", " ^ tail


and generate_parameter (p: parameter) =
    match p with
    ParameterOfTypeAndID (tp, id) ->  id


and generate_in_func_stc_list (lst: in_func_stc list) =
    match lst with
    | [] -> ""
    | head :: [] -> generate_in_func_stc head
    | head :: tail ->
        let head = generate_in_func_stc head in
        let tail = generate_in_func_stc_list tail in
        head ^ "\n" ^ tail


and generate_in_func_stc (stc: in_func_stc) =
    match stc with
    InFuncStcOfStmt stmt -> generate_stmt stmt


and generate_obj_def (o: obj_def) =
    match o with
    ObjDefOfIDAndStcList (objname, lst) -> 
        let lst = get_property_list_from_stc_list objname lst in
        let lst1 = generate_property_list lst in
        let lst2 = generate_property_assign_list lst in
        "function " ^ objname ^ "(" ^ lst1 ^ ")\n" ^
        "{\n" ^
        lst2 ^ "\n" ^
        "}\n"

    
and getTupleFirst (x, _, _) =
    x


and generate_property_list (lst) =
    match lst with
    [] -> ""
    | head :: [] -> getTupleFirst head
    | head :: tail ->
        getTupleFirst head ^ ", " ^ generate_property_list tail


and generate_property_assign_list (lst) =
    match lst with
    [] -> ""
    | head :: [] -> 
        let (name, barray, str) = head in
        (match barray with
        | true -> "this." ^ name ^ " =  new Array;"
        | _ -> 
            (if str = "" then
                "this." ^ name ^ " = " ^ name ^ ";"
            else
                "this." ^ name ^ " = new " ^ str ^ "();"))
    | head :: tail -> 
        let (name, barray, str) = head in
        let s = (match barray with
        | true -> "this." ^ name ^ " =  new Array;"
        | _ -> 
            (if str = "" then
                "this." ^ name ^ " = " ^ name ^ ";"
            else
                "this." ^ name ^ " = new " ^ str ^ "();"))
        in 
        s ^ "\n" ^ generate_property_assign_list tail


and get_property_list_from_stc_list (name: string) (lst: in_obj_stc list) =
    match lst with
    [] -> []
    | head :: [] -> get_property_from_stc name head
    | head :: tail ->
        let head = get_property_from_stc name head in
        let tail = get_property_list_from_stc_list name tail in
        head @ tail


and get_property_from_stc (name: string) (stc: in_obj_stc) =
    match stc with
    InObjStcOfVarDec v ->
        match v with 
        VarDecOfTypeAndDeclaratorList (tp, declst) ->
            let lst = get_property_from_dec_list name declst in
            objmap.map <- StringMap.add name lst objmap.map;
            let s = match tp with
            | CLTypeOfObjID s -> s
            | _ -> "" 
            in
            combineTupleList lst s


and getTupleEle (x, y) =
    (x, y)

and combineTupleList lst s =
    match lst with
    [] -> []
    | head :: [] -> 
        let (x,y) = getTupleEle head in
        [(x, y, s)]
    | head :: tail -> 
        let (x,y) = getTupleEle head in
        let tail = combineTupleList tail s in
        (x,y,s) :: tail


and get_property_from_dec_list (name: string) (lst: declarator list) =
    match lst with
    [] -> []
    | head :: [] -> 
        let (head, isarray) = get_declarator head in
        (*objmap.map <- StringMap.add head isarray objmap.map;*)
        [(head, isarray)]
    | head :: tail ->
        let (head, isarray) = get_declarator head in
        (*objmap.map <- StringMap.add head isarray objmap.map;*)
        let tail = get_property_from_dec_list name tail in
        (head, isarray) :: tail


and generate_var_dec (vc: var_dec) =
    match vc with
    VarDecOfTypeAndDeclaratorList (tp, declst) ->
        match tp with 
        CLTypeOfObjID s -> 
            let lst = get_declarator_list declst in
            generate_obj_declaration_list s lst
        | _ -> generate_declarator_list declst


and generate_obj_declaration_list (tp: string) (lst: string list) =
    match lst with
    | [] -> ""
    | head :: [] -> 
        "var " ^ head ^ " = new " ^ tp ^ "();\n"
    | head :: tail ->
        let s = generate_obj_new_array tp head in
        "var " ^ head ^ " = new " ^ tp ^ "();\n" ^
        generate_obj_declaration_list tp tail


and generate_obj_new_array (objname: string) (id: string) =
    let lst = StringMap.find objname objmap.map in
    let lst = List.fold_left f_fold [] lst in
    generate_new_array_list id lst


and f_fold lst (id, isarray) = 
    if isarray then
        id :: lst
    else 
        lst


and generate_new_array_list (id: string) (lst: string list) =
    match lst with
    [] -> ""
    | head :: [] -> id ^ "." ^ head ^ " = new Array();\n"
    | head :: tail -> 
        let head = id ^ "." ^ head ^ " = new Array();\n" in
        let tail = generate_new_array_list id tail in
        head ^ tail


and generate_stmt_list (lst: stmt list) =
    match lst with
    | [] -> "\n"
    | head :: [] -> generate_stmt head
    | head :: tail ->
        let head = generate_stmt head in
        let tail = generate_stmt_list tail in
        head ^ "\n" ^ tail


and get_declarator_list (lst: declarator list ) = 
    match lst with
    [] -> raise ( Failure("Empty Declarator List\n") )
    | head :: [] -> 
        let (h, _) = get_declarator head in
        [h]
    | head :: tail -> 
        let (h,_) = get_declarator head in
        let t = get_declarator_list tail in
        h :: t


and get_declarator (d: declarator) = 
    match d with
    DeclaratorOfID (s, _) -> (s, false)
    | DeclaratorOfArray (s, _) -> (s, true)
    | DeclaratorOfFixedArray (s, e, _) -> 
        (s, true)
    

and generate_declarator (d: declarator) =
    match d with
    DeclaratorOfID (s, _) -> "var " ^ s ^ ";"
    | DeclaratorOfArray (s, _) -> "var " ^ s ^ " = new Array();"
    | DeclaratorOfFixedArray (s, e, _) -> 
        "var " ^ s ^ " = new Array(" ^ generate_cst_expr e ^ ");"


and generate_declarator_list (declst: declarator list) =
    match declst with
    [] -> raise ( Failure("Empty Declarator List\n") )
    | head :: [] -> generate_declarator head
    | head :: tail -> 
        let h = generate_declarator head in
        let t = generate_declarator_list tail in
        h ^ "\n" ^ t


and generate_stmt (s: stmt) =
    match s with
    | StmtOfVarDec v -> 
        generate_var_dec v

    | StmtOfExprStmt e -> 
        generate_expr_stmt e

    | StmtOfSelectionStmt s -> 
        generate_selection_stmt s

    | StmtOfIterationStmt is -> 
        generate_iteration_stmt is

    | StmtOfJumpStmt js -> 
        generate_jump_stmt js

    | StmtOfDrawStmt ds -> 
        generate_draw_stmt ds


and generate_expr_stmt (e: expr_stmt) =
    match e with
    ExprStmtOfExpr e -> generate_expr e ^ ";\n"


and generate_draw_stmt (ds: draw_stmt) =
    match ds with
    DrawStmtOfFD e ->
        let s = generate_expr e in
        "_tmp = _direction * Math.PI / 180;\n" ^
        "_tmpx = Math.cos(_tmp) * ( " ^ s ^ " ) + _currx;\n" ^
        "_tmpy = Math.sin(_tmp) * ( " ^ s ^ " ) + _curry;\n" ^
        "if (_bDraw)\n" ^
        "  drawLine(_currx, _curry, _tmpx, _tmpy);\n" ^
        "_currx = _tmpx; _curry = _tmpy;\n"

    | DrawStmtOfBK e ->
        let s = generate_expr e in
        "_tmp = _direction * Math.PI / 180;\n" ^
        "_tmpx = _currx - Math.cos(_tmp) * ( " ^ s ^ " );\n" ^
        "_tmpy = _curry - Math.sin(_tmp) * ( " ^ s ^ " );\n" ^
        "if (_bDraw)\n" ^
        "  drawLine(_currx, _curry, _tmpx, _tmpy);\n" ^
        "_currx = _tmpx; curry = _tmpy;\n"

    | DrawStmtOfLF e ->
        let s = generate_expr e in
        "_direction -= " ^ s ^ ";\n"

    | DrawStmtOfRT e ->
        let s = generate_expr e in
        "_direction += " ^ s ^ ";\n"

    | DrawStmtOfReset -> 
        "_direction = 0;\n" ^
        "_currx = document.width / 2;\n" ^
        "_curry = document.height / 2;\n"

    | DrawStmtOfCls -> 
        "_context.clearRect(0, 0, _canvas.width, _canvas.height);\n"

    | DrawStmtOfPf -> 
        "_bDraw = !_bDraw;\n"

    | DrawStmtOfPd -> 
        "_bDraw = true;\n"

    | DrawStmtOfPu ->
        "_bDraw = false;\n"

    | DrawStmtOfWd e -> 
        "_context.lineWidth = " ^ generate_expr e ^ ";\n"

    | DrawStmtOfRGB (e1, e2, e3) ->
        let e1 = generate_expr e1 in
        let e2 = generate_expr e2 in
        let e3 = generate_expr e3 in
        "_r = " ^ e1 ^ ";\n" ^
        "_g = " ^ e2 ^ ";\n" ^
        "_b = " ^ e3 ^ ";\n" ^
        "_tmps = \"rgb(\" + _r.toString() + \", \" + _g.toString() + \", \" + _b.toString() + \")\"\n" ^
        "_context.strokeStyle = _tmps;\n"


and generate_selection_stmt (s: selection_stmt) =
    match s with
    SelectionStmtOfExprAndStmtList (e, lst) ->
        let e = generate_expr e in
        let lst = generate_stmt_list lst in
        "if (" ^ e ^ ")\n" ^
        "{\n" ^
        lst ^
        "}\n"
    
    | SelectionStmtOfExprAndStmtList2 (e, iflst, elselst) ->
        let e = generate_expr e in
        let iflst = generate_stmt_list iflst in
        let elselst = generate_stmt_list elselst in
        "if (" ^ e ^ ")\n" ^
        "{\n" ^
        iflst ^
        "}\n" ^
        "else\n" ^
        "{\n" ^
        elselst ^
        "}\n"


and generate_iteration_stmt (s: iteration_stmt) =
    match s with
    | IterationStmtOfExprAndStmtList (e, lst) ->
        let lst = generate_stmt_list lst in
        let e = generate_expr e in
        "_stack.push(_i);\n" ^
        "for (_i = 0; _i < (" ^ e ^ ")" ^ "; _i++)\n" ^
        "{\n" ^
        lst ^
        "}\n" ^
        "_i = _stack.pop();\n"
            
    | IterationStmtOfStmtList (lst) ->
        let lst = generate_stmt_list lst in 
        "while(true)\n" ^
        "{\n" ^
        lst ^
        "}\n"


and generate_jump_stmt (s: jump_stmt) =
    match s with
    JumpStmtOfGOON li-> "continue;\n"
    | JumpStmtOfBRK li-> "break;\n"
    | JumpStmtOfReturnExpr e -> 
        let e = generate_expr e in
        "return " ^ e ^ ";\n"
    | JumpStmtOfReturn li-> 
        "return"


and generate_expr (e: expr) =
    match e with
    | ExprOfPrimary (p, li) -> generate_primary p
    | ExprOfCstExpr (e, _) -> generate_cst_expr e
    | ExprOfAssign (l, v) ->
        let l = generate_lvalue l in
        let v = generate_expr v in
        l ^ " = " ^ v
    | ExprOfPlus (e1, e2) -> 
        generate_expr e1 ^ " + " ^ generate_expr e2
    | ExprOfMinus (e1, e2) -> 
        generate_expr e1 ^ " - " ^ generate_expr e2
    | ExprOfTimes (e1, e2) -> 
        generate_expr e1 ^ " * " ^ generate_expr e2
    | ExprOfDivide (e1, e2) -> 
        generate_expr e1 ^ " / " ^ generate_expr e2
    | ExprOfNeq (e1, e2) -> 
        generate_expr e1 ^ " != " ^ generate_expr e2
    | ExprOfEq (e1, e2) -> 
        generate_expr e1 ^ " == " ^ generate_expr e2
    | ExprOfLt (e1, e2) -> 
        generate_expr e1 ^ " < " ^ generate_expr e2
    | ExprOfGt (e1, e2) -> 
        generate_expr e1 ^ " > " ^ generate_expr e2
    | ExprOfLeq (e1, e2) -> 
        generate_expr e1 ^ " <= " ^ generate_expr e2
    | ExprOfGeq (e1, e2) -> 
        generate_expr e1 ^ " >= " ^ generate_expr e2
    | ExprOfUMinus (e1) -> 
        "(-" ^ generate_expr e1 ^ ")"
    | ExprOfNot e ->
        "!(" ^ generate_expr e ^ ")"
    | ExprOfFunction (id, lst, li) ->
        let lst = generate_argument_lst lst in
        id ^ "(" ^ lst ^ ")"


and generate_argument_lst (lst: expr list) =
    match lst with
    [] -> ""
    | head :: [] -> generate_expr head
    | head :: tail -> 
        let head = generate_expr head in
        let tail = generate_argument_lst tail in
        head ^ ", " ^ tail


and generate_primary (p: primary) =
    match p with
    | PrimaryOfIdentifier (id, _) -> generate_identifier id 
    | PrimaryOfArrayElement (id, e) -> generate_identifier id ^ "[" ^ generate_expr e ^ "]"


and generate_identifier (id: identifier) =
    match id with
    | IdentifierOfSimple (s, li) -> s
    | IdentifierOfObjElement (s, id, li) -> 
        let id = generate_identifier id in
        s ^ "." ^ id
        
          
and generate_cst_expr (e: cst_expr) =
    match e with
    CstExprOfInt i -> string_of_int i
    | CstExprOfReal f -> string_of_float f
    | CstExprOfTrue -> "true"
    | CstExprOfFalse -> "false"


and generate_lvalue (l: lvalue) =
    match l with
    LvalueOfIdentifier (id, li) -> generate_identifier id
    | LvalueOfArrayElement (id, e) -> 
        let id = generate_identifier id in
        id ^ "[" ^ generate_expr e ^ "]"
        

end

